/* This file is part of flowz.
   Copyright (C) 2006 Sergey Poznyakoff

   Flowz is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   Flowz is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with flowz; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
   MA 02110-1301 USA */

#include "flowz.h"

enum {
     VERSION_OPTION = 256,
     MAX_ERRORS_OPTION,
     MAX_MEMORY_OPTION,
     CHECKPOINT_OPTION
};

struct option longopt[] = {
     { "level", 1, NULL, 'l' },
     { "verbose", 0, NULL, 'v' },
     { "help", 0, NULL, 'h' },
     { "output", 1, NULL, 'o' },
     { "version", 0, NULL, VERSION_OPTION },
     { "max-errors", 1, NULL, MAX_ERRORS_OPTION },
     { "max-memory", 1, NULL, MAX_MEMORY_OPTION },
     { "checkpoint", 1, NULL, CHECKPOINT_OPTION },
     { NULL }
};

int level = 1;              /* Aggregation level */
int verbose = 0;            /* Verbosity level */
char *input = NULL;         /* Input file name */
char *output = NULL;        /* Output file name */ 
FILE *infile;               /* Input stream */
FILE *outfile;              /* Output stream */
size_t error_lines;         /* Number of erroneous lines encountered so far */
size_t max_error_lines;     /* Maximum allowed number of erroneous lines.
			       0 means no limit */
off_t max_memory_usage;     /* When memory usage reaches this number, dump
			       the collected data and clear the table */
size_t checkpoint_option;   /* Dump statistics each Nth line */
size_t lineno = 0;          /* Number of the input line currently being
			       processed */
 
char *program_name;         /* Executable name */

struct {                    /* Dispatcher table */ 
     compress_fn fun;       /* Aggregation function */
     char *header;          /* Header to be output before it. It must be
 			       a comma-separated list of field names, since
			       it is also used to determine maximum number
			       of fields for the next aggregation method */
} compress[] = {
     { NULL,
       "unix_secs,doctets,srcaddr,dstaddr,tos,src_mask,dst_mask,src_as,dst_as,src_tag,dst_tag" },
     { level_1,
       "first-last,doctets,srcaddr,dstaddr,tos,src_mask,dst_mask,src_as,dst_as,src_tag,dst_tag"},
     { level_2,
       "first-last,doctets,srcbase,dstbase,tos,src_mask,dst_mask,src_as,dst_as,src_tag,dst_tag" },
     { level_3,
       "first-last,doctets,tos,src_as,dst_as,src_tag,dst_tag" },
};

int max_level = sizeof(compress) / sizeof(compress[0]);

char *header_str; /* Points to the header line to be used */

char usage[] = "\n\
OPTIONS are:\n\
\n\
  --checkpoint=NUMBER        print line number and hash statistics each\n\
                             NUMBERth line; with -vv dumps full hash stats\n\
  -l, --level=NUMBER         aggregation level number (default 1, max 3)\n\
  --max-errors=NUMBER        tolerate at most NUMBER errors\n\
  --max-memory=SIZE          limit memory consumption to SIZE bytes; use with\n\
                             caution.\n\
  -o, --output=FILE          set output file name (default - stdout)\n\
  -v, --verbose              print additional statistics at the end of the run\n\
\n\
  -h, --help                 print this help message\n\
  -v, --version              print program version and exit\n";

void
help()
{
     printf("%s - flow statistics aggregator\n", program_name);
     printf("Usage: %s [OPTIONS] FILE\n", program_name);
     printf("%s", usage);
     exit(EXIT_SUCCESS);
}

void
throw_error()
{
     ++error_lines;
     if (max_error_lines && error_lines == max_error_lines)
	  error(EXIT_FAILURE, 0, "Too many errors");
}

/* For each line from the input, parse it, check if it has NFIELDS comma-
   separated fields and pass it to the handler function FN */
int
read_file(compress_fn fn, size_t nfields)
{
     char *buf = NULL;
     size_t size = 0;
     char **toks;
     size_t i;
     int rc = EXIT_SUCCESS;
     
     toks = xcalloc(nfields, sizeof toks[0]);
     while (getline(&buf, &size, infile) > 0) {
	  char *p;

	  lineno++;
	  if (checkpoint_option && (lineno % checkpoint_option) == 0) {
	       fprintf(stderr, "* Line %lu, entries %lu\n",
		       (unsigned long) lineno, table_n_entries());
	       if (verbose > 2) 
		    table_stat(stderr);
	  }

	  for (p = buf; *p && isspace(*p); p++)
	       ;
	  if (*p == 0 || *p == '#')
	       continue;

	  for (i = 0, p = strtok(buf, ","); p; p = strtok(NULL, ",\n")) {
	       if (i == nfields) {
		    error(0, 0, "%s:%lu: too many fields",
			  input, (unsigned long) lineno);
		    throw_error();
		    continue;
	       }
	       toks[i++] = p;
	  }

	  if (i != nfields) {
	       error(0, 0, "%s:%lu: not enough fields",
		     input, (unsigned long) lineno);
	       throw_error();
	       continue;
	  }
	  if (fn(toks, nfields)) 
	       throw_error();
     }
     return error_lines == 0 ? EXIT_SUCCESS : EXIT_MILDFAIL;
}

/* Create a key by concatenating tokens from TOKS (separated by commas).
   MAP contains indices of tokens to be used. MAPSIZE is the number of
   entries in MAP. */
char *
make_key(unsigned char *map, int mapsize, char **toks)
{
     static char *buf = NULL;
     static size_t size = 0;
     int i;
     size_t len = mapsize - 1;
     char *p;
     
     for (i = 0; i < mapsize; i++) 
	  len += strlen(toks[map[i]]);

     len++;

     if (len > size) {
	  size = len;
	  buf = xrealloc(buf, size);
     }

     for (i = 0, p = buf; ;) {
	  p = stpcpy(p, toks[map[i]]);
	  if (++i < mapsize)
	       *p++ = ',';
	  else
	       break;
     }
     return buf;
}

/* Count number of fields in comma-separated list STR */
size_t
numfields(const char *str)
{
     size_t nf = 1;
     for (; *str; str++)
	  if (*str == ',')
	       nf++;
     return nf;
}


int
main(int argc, char **argv)
{
     int rc;
     char *p;
     
     program_name = argv[0];
     
     while ((rc = getopt_long(argc, argv, "hl:o:v", longopt, NULL)) != EOF) {
	  switch (rc) {
	  case 'l':
	       level = strtoul(optarg, &p, 0);
	       if (*p)
		    error(EXIT_FAILURE, 0,
			  "Compression level must be numeric");
	       if (level > max_level || level == 0)
		    error(EXIT_FAILURE, 0, "Unknown compression level");
	       break;

	  case 'h':
	       help();
	       break;

	  case 'o':
	       output = optarg;
	       break;
	       
	  case 'v':
	       verbose++;
	       break;

	  case CHECKPOINT_OPTION:
	       checkpoint_option = strtoul(optarg, &p, 0);
	       if (*p) 
		    error(EXIT_FAILURE, 0,
			  "Checkpoint must be numeric");
	       break;
	       
	  case MAX_ERRORS_OPTION:
	       max_error_lines = strtoul(optarg, &p, 0);
	       if (*p) 
		    error(EXIT_FAILURE, 0,
			  "Maximum error count must be numeric");
	       break;

	  case MAX_MEMORY_OPTION:
	  {
	       uintmax_t n;
	       
	       n = strtoumax(optarg, &p, 0);
	       if (errno == ERANGE) 
		    error(EXIT_FAILURE, 0,
			  "Maximum memory usage is out of allowed range");
	       if (*p) {
		    uintmax_t m;
		    
		    if (p[1] && ((p[1] != 'b' || p[1] != 'B') || p[2]))
			 error(EXIT_FAILURE, 0, "Unknown suffix: %s", p);
		    switch (*p) {
		    case 'k':
		    case 'K':
			 m = n*1024;
			 break;
			 
		    case 'm':
		    case 'M':
			 m = n*1024*1024;
			 break;
			 
		    case 'g':
		    case 'G':
			 m = n*1024*1024*1024;
			 break;
			 
		    default:
			 error(EXIT_FAILURE, 0, "Unknown suffix: %s", p);
			 break;
		    }

		    if (m < n) 
			 error(EXIT_FAILURE, 0,
			       "Maximum memory usage is out of allowed range");
		    n = m;
	       }

	       if (n > TYPE_MAXIMUM(off_t))
		    error(EXIT_FAILURE, 0,
			  "Maximum memory usage is out of allowed range");
	       max_memory_usage = n;
	       break;
	  }
		    
	  case VERSION_OPTION:
	       printf("%s\n", PACKAGE_STRING);
	       exit(0);
	       
	  default:
	       exit(EXIT_FAILURE);
	  }
     }

     if (optind == argc)  
	  error(EXIT_FAILURE, 0,
		"Not enough arguments. Try `%s --help' for more info",
		program_name);

     input = argv[optind];
     if (strcmp(input, "-") == 0) {
	  infile = stdin;
	  input = "<stdin>";
     } else {
	  infile = fopen(argv[optind], "r");
	  if (!infile)
	       error(EXIT_FAILURE, errno, "cannot open input file %s: %s",
		     argv[optind], strerror(errno));
     }
     
     if (output) {
	  outfile = fopen(output, "w");
	  if (!outfile)
	       error(EXIT_FAILURE, errno, "cannot open output file %s: %s",
		     output, strerror(errno));
     } else {
	  output = "<stdout>";
	  outfile = stdout;
     }

     table_init();
     header_str = compress[level].header;
     rc = read_file(compress[level].fun,
		    numfields(compress[level - 1].header));
     dump_table(outfile, header_str);
     if (verbose)
	  table_stat(stdout);
     fclose(outfile);
     fclose(infile);
     return rc;
}


     
	  
	  
	  


